/************************************************************/
/* For the purpose of this program see the function "usage" */
/*                                                          */
/* ABSOLUTE NO WARRANTY!                                    */
/*                                                          */
/* Sun Mar  8 21:49:12 MET 1998                             */
/* Author: J. Quade                       quade@softing.com */
/* (c) 1998 by the GPL                                      */
/************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "zipsecure.h"

/*extern void ResetFileName(); */
/*                             */
/* V1.5: bug fixing            */

static char *Version="V1.5";
static char *Author="quade@amavis.org";
static char *Date="Fri May 26 15:51:10 CEST 2000";

/* attempt to hack in big endian support c.mock@xsoft.co.at 1999 */
/* tested on IBM AIX 4.2.x */
/* define BIGENDIAN to activate */
/*#define DEBUG*/
# ifdef BIGENDIAN
#  define SWAPSHORT(x) ((((x) & 0x00ff) << 8) | (((x) & 0xff00) >> 8))
#  define SWAPLONG(x) ((((x) & 0xff) << 24) | (((x) & 0xff00) << 8) | \
		     (((x) & 0xff0000) >> 8) | (((x) & 0xff000000) >> 24))
# else
#  define SWAPSHORT(x) (x)
#  define SWAPLONG(x) (x)
# endif /* !BIGENDIAN */


/*
Zip-Format

for details see:
	http://www.cdrom.com/pub/infozip/doc/appnote*iz.zip

Structure of a zip-file _SEEMS_ to be the following:
A zip file contains a set of entries. Three entries are distinguished:
- A local header entry (local file)
- A central header entry
- An End of Archive entry
An entry is identified by the first 4 bytes:

           Byte  0    1    2    3
central header: 0x50 0x4b 0x01 0x02
local header:   0x50 0x4b 0x03 0x04
end of archive: 0x50 0x4b 0x05 0x06

To read a complete entry, one has
1. to read the header (fixed length to read)
2. to read the filename (length to read depends on the name length)
3. to read the extra field (length depends)
4. to read the compressed data (in case of a local header)

Each central directory entry stores the relative offset to
the begin of the file.

The end of central dir record contains the size of
the central directory.

The general_purpose_flag _SEEMS_ to have the following meaning:
	bit 0	- encrypted or not
	bit 1	- sliding window (for deflation) 0=4k, 1=8k
	bit 2	- deflation type
	bit 3
	bit 4
	bit 5
	bit 6
	bit 7	- extended local header

*/

#if SIZEOF_UNSIGNED_SHORT==2
typedef unsigned short u16;
#endif

#if SIZEOF_UNSIGNED_SHORT==4
typedef unsigned short u32;
#elif SIZEOF_UNSIGNED_INT==4
typedef unsigned int u32;
#elif SIZEOF_UNSIGNED_LONG==4
typedef unsigned long u32;
#endif


typedef enum { Start, ReadingFiles, ReadingCentralDir } mode;

typedef struct _localheader {
	u16 Magic;
	u16 VersionNeeded;
	u16 flag;
	u16 CompressionMethod;
	u16 FileModificationTime;
	u16 FileModificationDate;
	u32 crc;
	u32 CompressedSize;
	u32 FileLength; /* uncompressed */
	u16 FileNameLength;
	u16 ExtraFieldLength;
} __attribute__ ((packed)) LocalZipHeader;

typedef struct _zipheader {
	u16 Magic;
	u16 VersionNeeded;
	u16 flag;
	u16 CompressionMethod;
	u16 FileModificationTime;
	u16 FileModificationDate;
	u32 crc;
	u32 CompressedSize;
	u32 FileLength; /* uncompressed */
	u16 FileNameLength;
	u16 ExtraFieldLength;
	u16 FileCommentLength;
	u16 DiskNrStart;
	u16 InternalFileAttr;
	u32 ExternalFileAttr;
	u32 RelativeOffsetLocalHeader;
} __attribute__ ((packed)) ZipHeader;

typedef struct _EndOfCentralDir {
	u16 EndOfCentralDirMagic;
	u16 NumberThisDisk;
	u16 NumDiskWithStartCentralDir;
	u16 NumEntriesCentralDirThsDisk;
	u16 TotalEntriesCentralDir;
	u32 SizeCentralDir;
	u32 OffsetStartCentralDir;
	u16 ZipfileCommentLength;
} __attribute__ ((packed)) EndOfCentralDir;

static int FileNameNumber=0;
static int ReducedLength;

void ResetFileName()
{
	FileNameNumber=0;
}

static void LocalHeaderOut( FILE *fp, LocalZipHeader *Header, char *name)
{
	struct _zipheader *hdr = (struct _zipheader *)Header;

	fprintf(fp,"%s\n", name );
	fprintf(fp,"\tMagic: 0x%x\n", SWAPLONG(Header->Magic) );
	fprintf(fp,"\tVersionNeeded: 0x%x\n", SWAPSHORT(Header->VersionNeeded) );
	fprintf(fp,"\tflag: 0x%x\n", SWAPSHORT(Header->flag) );
	fprintf(fp,"\tCompressionMethod: 0x%x\n",
		SWAPSHORT(Header->CompressionMethod) );
	fprintf(fp,"\tcrc: 0x%lx\n", SWAPLONG(Header->crc) );
	fprintf(fp,"\tcompressed/uncompressed size: %ld/%ld\n",
		SWAPLONG(Header->CompressedSize), SWAPLONG(Header->FileLength) );
	fprintf(fp,"\tFileNameLength: 0x%x\n",
		SWAPSHORT(Header->FileNameLength) );
	fprintf(fp, "\tExtraFieldLength: %d\n",
		SWAPSHORT(Header->ExtraFieldLength) );
	fprintf(fp, "\tMod. Date/Time: %d/%d\n",
		SWAPSHORT(Header->FileModificationDate),
		SWAPSHORT(Header->FileModificationTime) );
	fprintf(fp, "\tOffset: %ld (%lx)\n",
		SWAPLONG(hdr->RelativeOffsetLocalHeader),
		SWAPLONG(hdr->RelativeOffsetLocalHeader) );
}

static char *GetNextFileName( char *OldFileName )
{
	char buffer[256], *Extension;

	// find extension
	for( Extension=OldFileName+strlen(OldFileName); Extension>OldFileName
		&& *Extension != '.'; Extension-- );
	// We check the Extension length just in order to avoid an buffer-
	// overflow on variable "buffer"
	if( *Extension=='.' && strlen(Extension)<100) {
		snprintf(buffer, sizeof(buffer),"y%d.%d%s", getpid(),
				FileNameNumber++, Extension );
	} else {
		snprintf(buffer,sizeof(buffer),"y%d.%d", getpid(), FileNameNumber++ );
	}
	return( strdup(buffer) );
}

static long LoadLocalHeader( FILE *fpin, FILE *fpout, FILE *fperr,
	LocalZipHeader *Header )
{
	char *NamePtr, buffer[256], *NewName;
	int fnl, BytesRead, i, MaxSize, ToRead;
	long BytesWritten=0;

	fnl = SWAPSHORT( Header->FileNameLength );
	NamePtr = (char *)malloc( fnl+1 );
	BytesRead = fread( NamePtr, fnl, 1, fpin );
	NamePtr[fnl] = '\0';
	NewName = GetNextFileName( NamePtr );
	fnl = strlen(NewName);
	Header->FileNameLength = SWAPSHORT( fnl );
	if( fperr )
		fprintf(fperr,"changing \"%s\" to \"%s\"(%d)\n", NamePtr, NewName,fnl);
	fwrite( Header, sizeof(LocalZipHeader), 1, fpout );
	fwrite( NewName, fnl, 1, fpout );
	BytesWritten += sizeof(LocalZipHeader);
	BytesWritten += fnl;
	ReducedLength -= strlen(NamePtr) - fnl;
	BytesRead = fread( buffer, 1, SWAPSHORT(Header->ExtraFieldLength), fpin );
	fwrite( buffer, 1, BytesRead, fpout );
	BytesWritten += BytesRead;
	i=0;
	for( ToRead=SWAPLONG(Header->CompressedSize); ToRead; ToRead-=BytesRead ) {
		MaxSize = (ToRead>sizeof(buffer)) ? sizeof(buffer):ToRead;
		BytesRead = fread( buffer, 1, MaxSize, fpin );
		fwrite( buffer, 1, BytesRead, fpout );
		BytesWritten += BytesRead;
		i+=BytesRead;
		if( !BytesRead ) {
			if( fperr )
				fprintf(fperr,"Archive corrupted (EOF) (ToRead=%d)\n",ToRead);
			BytesWritten=0;
			break;
		}
	}
#ifdef DEBUG
	if( BytesWritten )
		LocalHeaderOut( stderr, Header, NamePtr );
#endif
	if( NamePtr )
		free( NamePtr );
	if( NewName )
		free( NewName );
	return( BytesWritten );
}

static long LoadCentralDirEntry( FILE *fpin, FILE *fpout, FILE *fperr,
	ZipHeader *Header )
{
	char *NamePtr, buffer[256], *NewName;
	int ToCopy, fnl, BytesRead, i, MaxSize, ToRead, CurLengthToReduce;
	static int SumLengthToReduce;
	long BytesWritten=0;
	

	if( FileNameNumber==0 )
		SumLengthToReduce = 0;
	fnl = SWAPSHORT(Header->FileNameLength);
	NamePtr = (char *)malloc( fnl+1 );
	BytesRead = fread( NamePtr, fnl, 1, fpin );
	NamePtr[fnl] = '\0';
	NewName = GetNextFileName( NamePtr );
	fnl = strlen( NewName );
	Header->FileNameLength = SWAPSHORT( fnl );
	CurLengthToReduce = strlen( NamePtr ) - fnl;
#ifdef DEBUG
	if( fperr )
		fprintf(fperr, "Offset=%ld (%lx) - ",
			SWAPLONG(Header->RelativeOffsetLocalHeader),
			SWAPLONG(Header->RelativeOffsetLocalHeader) );
#endif
	Header->RelativeOffsetLocalHeader = 
		SWAPLONG( SWAPLONG(Header->RelativeOffsetLocalHeader)
			-SumLengthToReduce);
	SumLengthToReduce += CurLengthToReduce;
	ReducedLength -= CurLengthToReduce;
	if( fperr ) {
#ifdef DEBUG
		fprintf(fperr, "%ld (%lx)\n",
			Header->RelativeOffsetLocalHeader,
			Header->RelativeOffsetLocalHeader );
#endif
		fprintf(fperr,"changing \"%s\" to \"%s\"\n",NamePtr, NewName );
	}
	fwrite( (void *)Header, 1, sizeof(ZipHeader), fpout );
	fwrite( NewName, fnl, 1, fpout );
	BytesWritten += sizeof(ZipHeader);
	BytesWritten += fnl;
	for( ToCopy=SWAPSHORT(Header->ExtraFieldLength); ToCopy; ToCopy-=i ) {
		i = fread(buffer,1,ToCopy>sizeof(buffer)?sizeof(buffer):ToCopy,fpin);
		fwrite( buffer, 1, i, fpout );
	}
	BytesWritten += SWAPSHORT(Header->ExtraFieldLength);
#ifdef DEBUG
	LocalHeaderOut( stderr, (LocalZipHeader *)Header, NamePtr );
#endif
	if( NamePtr )
		free( NamePtr );
	if( NewName )
		free( NewName );
	return( BytesWritten );
}

static void Usage( char *Name )
{
	fprintf(stdout,"usage: %s [-q] [-v]\n", Name);
	fprintf(stdout,"\t-q: quiet\n");
	fprintf(stdout,"\t-v: version\n");
	fprintf(stdout,"Program reads a zip file from stdin, removes any pathes\n");
	fprintf(stdout,"of a contained file and changes the name of the\n");
	fprintf(stdout,"file to a new file name. The new name starts with a\n");
	fprintf(stdout,"'z', followed by the process ID and a sequence number.\n");
	fprintf(stdout,"If any extension in the original name was present,\n");
	fprintf(stdout,"it is also appended to the new name.\n");
}

int main( int argc, char **argv )
{
	FILE *fpin=stdin, *fpout=stdout, *fperr=stderr;
	ZipHeader *CentralHeader;
	LocalZipHeader *Header;
	EndOfCentralDir *EndOfZip;
	char buffer[256], *ptr;
	int i,end=0, BytesRead, ToRead;
	mode Mode=Start;
	long BytesWritten=0L;

#ifdef NODEBUG
	fperr = NULL; // no debugs
#endif
	
	if( sizeof(u16)!=2 || sizeof(u32)!=4 ) {
		if( fperr ) {
			fprintf(fperr,"u32 is not 4 Bytes or/and u16 not 2 Bytes!\n");
			fprintf(fperr,"Program not suitable for this machine!\n");
		}
		exit( -1 );
	}
	*((unsigned short *)buffer) = 0x1234;
	switch( buffer[0] ) {
	case 0x12: // big endian machine
#ifdef DEBUG
		fprintf(fperr,"Big Endian\n");
#endif
		*((unsigned short *)buffer) = SWAPSHORT(0x1234);
		if( buffer[0] == 0x34 ) {
			if( fperr )
				fprintf(fperr, "Program is compiled for a little endian "
						"machine,\nbut machine is big endian!\n" );
		}
		break;
	case 0x34: // little endian machine
#ifdef DEBUG
		fprintf(fperr,"Little Endian\n");
#endif
		*((unsigned short *)buffer) = SWAPSHORT(0x1234);
		if( buffer[0] == 0x12 ) {
			if( fperr )
				fprintf(fperr, "Program is compiled for a big endian "
						"machine,\nbut machine is little endian!\n" );
		}
		break;
	default:
		fprintf(fperr,"unknown machine type!\n");
		break;
	}
#ifdef BIGENDIAN
#ifdef DEBUG
	if( fperr )
		fprintf(fperr,"swap test: 0x12345678 == %08x, 0x1234 == %04x\n",
			SWAPLONG(0x12345678), SWAPSHORT((unsigned short)0x1234) );
#endif
#endif
	if( argc>3 ) {
		Usage( argv[0] );
		exit( -2 );
	}
	for( i=1; i<argc; i++ ) {
		if( strcmp(argv[i], "-q")==0 ) {
			fperr = NULL;
		} else if( strcmp(argv[i], "-v")==0 ) {
			fprintf(fpout, "%s: \"%s\" from \"%s\"\n",
					argv[0], Version, Date );
			fprintf(fpout, "Copyright 1999, J. Quade. This is free software;\n"
				"see the GNU General Public Licence version 2 or later\n"
				"for copying conditions. There is NO warranty.\n");
			fprintf(fpout, "%s\n", Author );
			exit( 0 );
		} else {
			Usage( argv[0] );
			exit( -3 );
		}
	}
	Header = (LocalZipHeader *)buffer;
	CentralHeader = (ZipHeader *)(buffer+2);
	EndOfZip = (EndOfCentralDir *)buffer;
	ResetFileName();
	while( !end ) {
		BytesRead = fread( buffer, 1, 1, fpin ); // Read the PK magic
		if( !BytesRead ) {
			end=1;
			break;
		}
		fwrite( buffer, 1, BytesRead, fpout );
		BytesWritten+=BytesRead;
		if( buffer[0]!=0x50 )
			continue;
		BytesRead = fread( buffer, 1, 1, fpin ); // Read the PK magic
		if( !BytesRead ) {
			end=1;
			break;
		}
		fwrite( buffer, 1, BytesRead, fpout );
		BytesWritten+=BytesRead;
		if( buffer[0]!=0x4b ) {
#ifdef DEBUG
			//if( fperr )
			//	fprintf(fperr,"NO MAGIC FOUND!\n");
#endif
			continue;
		}
#ifdef DEBUG
		if( fperr )
			fprintf(fperr,"Magic at offset 0x%lx - \n", ftell(fpin) );
#endif
		BytesRead = fread( buffer, 1, 2, fpin );
		BytesRead += fread( &buffer[2], 1, sizeof(LocalZipHeader)-2, fpin );
		if( buffer[0]=='\1' && buffer[1]=='\2' ) { // central dir
			//if( buffer[2] != 0x14 ) {
			//	fprintf(fperr,"Wrong Flags 0x%x?\n", buffer[2]);
			//	continue;
			//}
			if( Mode != ReadingCentralDir )
				ResetFileName();
			Mode = ReadingCentralDir;
			if( fperr )
				fprintf(fperr,"CentralDir: " );
			ptr = buffer+sizeof(LocalZipHeader);
			ToRead = sizeof(ZipHeader)-sizeof(LocalZipHeader)+2;
			BytesRead = fread( ptr, ToRead, 1, fpin );
			fwrite( buffer, 1, 2, fpout );
			BytesWritten+=2;
			BytesWritten+=LoadCentralDirEntry(fpin,fpout,fperr,CentralHeader);
		} else if( buffer[0]=='\3' && buffer[1]=='\4' ) { // local header
			// check, whether the entry it is a real entry or not
			if( buffer[2]&0x60 ) { // bits 5-6 have to be 0
				if( fperr ) {
					fprintf(fperr, "general_purpose_flags: 0x%x\n", buffer[2] );
					fprintf(fperr,"Wrong Flags 0x%x?\n", buffer[2]);
				}
				fwrite( buffer, sizeof(LocalZipHeader), 1, fpout );
				BytesWritten += sizeof(LocalZipHeader);
				continue;
			}
			//BytesRead=fread( &buffer[2], 1, sizeof(LocalZipHeader)-2, fpin );
			if( Mode != ReadingFiles )
				ResetFileName();
			Mode = ReadingFiles;
			if( fperr )
				fprintf(fperr,"File: " );
			BytesWritten += LoadLocalHeader( fpin, fpout, fperr, Header );
		} else if( buffer[0]=='\5' && buffer[1]=='\6' ) {
			EndOfZip->SizeCentralDir = 
				SWAPLONG(SWAPLONG(EndOfZip->SizeCentralDir)+ReducedLength/2);
			EndOfZip->OffsetStartCentralDir =
				SWAPLONG(SWAPLONG(EndOfZip->OffsetStartCentralDir)
				+ReducedLength/2);
			fwrite( buffer, 1, BytesRead, fpout );
			BytesWritten += BytesRead;
		} else {
			fprintf(stderr,"unknown entry type %d/%d\n", buffer[0],buffer[1]);
		}
	}
	if( fperr )
		fprintf(fperr,"Reduced Length: %d\n", ReducedLength );
	while( BytesRead = fread( buffer, 1, sizeof(buffer), fpin ) ) {
		fwrite( buffer, 1, BytesRead, fpout );
		BytesWritten += BytesRead;
	}
}
